/********************************************************************/
/** @file Previeer.cpp 
 * Description: Camera image canvas.				    *
 ********************************************************************/
#include <math.h>
#include <stdlib.h>

#include <wx/wxprec.h>
#ifndef WX_PRECOMP
  #include <wx/wx.h>
#endif
#include <wx/cursor.h>
#include <wx/dcbuffer.h>
#include <wx/rawbmp.h>

#include "Previewer.h"
#include "RT_EXEC.h"
#include "../library/hardware/OsSupp.h"
#include "../library/images/ras_comp.h"

#define DEBUG   0  

/**
*  ----------------------------------------------------------------------------
*  event tables and other macros for wxWidgets
*  ----------------------------------------------------------------------------
*
*  the event tables connect the wxWidgets events with the functions (event
*  handlers) which process them. It can be also done at run-time, but for the
*  simple menu events like this the static method is much simpler.
*
*/
BEGIN_EVENT_TABLE(CameraPreview, wxPanel) //ScrolledWindow)
    EVT_MOUSE_EVENTS(       CameraPreview::OnMouseMove)
    EVT_PAINT(              CameraPreview::OnPaint)
    EVT_ERASE_BACKGROUND(   CameraPreview::OnEraseBackground)
    EVT_TIMER(ID_TIMER_PREVIEW, CameraPreview::OnPreviewImageTimer)
END_EVENT_TABLE()


/**
 *  Constructor.
 *  @param[in] parent The parent window object.  */
CameraPreview::CameraPreview(wxWindow *parent, CameraTool *NewCT): 
	      wxPanel(parent, wxID_ANY, wxDefaultPosition, wxDefaultSize, 
                 wxRAISED_BORDER|wxTAB_TRAVERSAL|wxNO_FULL_REPAINT_ON_RESIZE|wxCLIP_CHILDREN),
              DEFAULT_BGROUND(_T("BLACK")),
              DEFAULT_TCIRCLE_COLOR(_T("RED")),
              DEFAULT_VCIRCLE_COLOR(_T("YELLOW")),	//WHITE
              DEFAULT_GRID_LINE(_T("WHITE"), 2, wxDOT),
              DIVISOR_RATIO(20),              
              ViewMode(RAW_COLOR),
	      filter(0),
              vflip(0),
              hflip(0),
	      rectangle(0),
	      CT(NewCT)
{
  ImageChanged = false;
  bit = NULL;

  ImageBuf = CreateCompRaster2DRGB(1,1,8);
  FilterBuf = CreateCompRaster2DRGB(1,1,8);

	/* Initialise circles first of all */    
  targetCircle = new CircRTShape(&MRT.ReqX,&MRT.ReqY);
  targetCircle->PaintCross=true;
  videoCircle = new CircRTShape(MRT.VideoX,MRT.VideoY);
  if(CT)
    {
    videoCircle->xPos = CT->GetVideoX();
    videoCircle->yPos = CT->GetVideoY();
    }
    
  targetCircle->Tx=videoCircle->Tx=&Tx;    
  targetCircle->color=DEFAULT_TCIRCLE_COLOR;
  targetCircle->style=wxTRANSPARENT;        
  videoCircle->color = DEFAULT_VCIRCLE_COLOR;
  videoCircle->style = wxTRANSPARENT;	//wxSOLID;
  targetCircle->Radius=videoCircle->Radius = 2.0/DIVISOR_RATIO;

  backgroundColor = wxColour(DEFAULT_BGROUND);    

  wxSize size = this->GetSize(); 
  Tx.X.End = -1-videoCircle->Radius/2;
  Tx.X.Start = +1+videoCircle->Radius/2;
  Tx.Y.End = -1-videoCircle->Radius/2;
  Tx.Y.Start = +1+videoCircle->Radius/2;

  Tx.X.CanvasEnd=size.GetX()+size.GetWidth(); //wxSize size;
  Tx.X.CanvasStart=size.GetX(); // + radius/2;
  Tx.Y.CanvasEnd=size.GetY()+size.GetHeight();
  Tx.Y.CanvasStart=size.GetY();

  GrY.Tx=GrX.Tx=&Tx;
  GrX.Ticks=GrY.Ticks=GRID_LINE_NUMBER;
  GrX.LinePen=GrY.LinePen=wxPen(DEFAULT_GRID_LINE);

      // Initializes all available image handlers
  wxInitAllImageHandlers();
  
  wxCursor down_cursor;
  if(WxLoadCursor(&down_cursor,_T("needle.cur")))
    {      
    this->SetCursor(wxCursor(down_cursor));
    }
  else  
    {
#ifdef DEBUG    
        wxString str;
        str.Printf( wxT("Error occured camera cursor loading!"));
        wxMessageBox(str, _T("CameraPreviewCameraPreview::constructor()"), wxOK | wxICON_INFORMATION, this);
#endif
    }
    
  m_dragMode = TEST_DRAG_NONE;        
}


CameraPreview::~CameraPreview()
{
  if(targetCircle) {delete targetCircle;targetCircle=NULL;}
  if(videoCircle) {delete videoCircle;videoCircle=NULL;}
  if(ImageBuf) {delete ImageBuf;ImageBuf=NULL;}
  if(FilterBuf) {delete FilterBuf;FilterBuf=NULL;}
  if(bit!=NULL) {delete bit;bit=NULL;}
}


void CameraPreview::SetImage(Raster2DAbstract * pRas)
{
  Raster2DAbstract *pRasDest = (filter==0) ? ImageBuf : FilterBuf;
  if(pRasDest)
  {
    if(pRas == NULL)
    {
      pRasDest->Erase();
    }
    else
    {
      pRasDest->Allocate2D(pRas->Size1D,pRas->Size2D);
      for(unsigned y=0; y<pRas->Size2D; y++)
      {
        pRasDest->Set(y,*pRas->GetRowRaster(y));
      }
    }
  }
  ImageChanged = true;
}



/** Periodical timer for controll */
void CameraPreview::OnPreviewImageTimer(wxTimerEvent& WXUNUSED(event)) 
{
    // refresh the picture panel
  Refresh();
}


void CameraPreview::OnCameraChange(wxCommandEvent & event)
{
  if(videoCircle)
    {
    videoCircle->xPos = MRT.VideoX;
    videoCircle->yPos = MRT.VideoY;
    }
}


/**
 *  When the size is changed, it is called and then it refreshs the screen and sets the layouts   
 * 
 *  @param[in] event Unused event
 *  \warning If the drawing depends on the size of the window, 
 *  \warning clearing the DC explicitly may be needed and repainting the whole window. 
 *  \warning In which case, calling <CODE>wxWindow::Refresh()</CODE> may be needed to invalidate the entire window. */
void CameraPreview::OnEraseBackground(wxEraseEvent& WXUNUSED(event))
{
   // do nothing
}


/** Set all values to default and repaint camera canvas. */
void CameraPreview::Reset()
{
  MRT.ReqX.SetDouble(0);
  MRT.ReqY.SetDouble(0);  

  Refresh(false);	//repaint a Ball&Plate after reset
}


/*
double GetInterval(void)
    {
      static LARGE_INTEGER pFrequency = {0};
      LARGE_INTEGER CurrentTime;

      if(pFrequency.QuadPart==0)
        if(QueryPerformanceFrequency(&pFrequency)==0)
          pFrequency.QuadPart = 1;

      if(pFrequency.QuadPart <= 1) return GetTickCount();

      QueryPerformanceCounter(&CurrentTime);
      return (1000.0*CurrentTime.QuadPart) / pFrequency.QuadPart;
    }
*/

wxBitmap *Raster2Bitmap(Raster2DAbstract *pRas, wxBitmap *pBitmap, unsigned width, unsigned height)
{  
  if(pRas==NULL || pRas->Size1D<=0 || pRas->Size2D<=0)
  {
    if(pBitmap!=NULL) delete pBitmap;
    return NULL;
  }

  if(pBitmap==NULL) pBitmap=new wxBitmap(width,height,24);
  else
  {
    if(pBitmap->GetWidth()!=width || pBitmap->GetHeight()!=height)
    {
      delete pBitmap;
      pBitmap = new wxBitmap(width,height,24);
    }
  }

  typedef wxPixelData<wxBitmap, wxNativePixelFormat> PixelData;
  PixelData data(*pBitmap);
  if(!data)
    {
    if(pBitmap!=NULL) delete pBitmap;
    //double T1=GetInterval();
    wxImage img(pRas->Size1D, pRas->Size2D, (unsigned char *)(pRas->GetRow(0)), true);
    img = img.Rescale(width, height);	// we need to resize the image before.
    pBitmap = new wxBitmap(img);
    //T1 = GetInterval() - T1;
    return pBitmap;
    }

  //double T1=GetInterval();
  PixelData::Iterator p(data);
  p.Offset(data, 0, 0);

  const unsigned YScale = (pRas->Size2D*0x10000) / height;
  const unsigned XScale = (pRas->Size1D*0x10000) / width;
  for(int y=0; y<height; ++y)
  {
    PixelData::Iterator rowStart = p;
    if(pRas->GetPlanes()==24)
    {
      BYTE *Raw = (BYTE*)(pRas->GetRow((y*YScale)>>16));
      if(Raw==NULL) continue;
      int xx = 0;
      for(int x=0; x<width; ++x, ++p)
      {      
        p.Red() = *Raw;
        p.Green() = Raw[1];
        p.Blue() = Raw[2];

        xx += XScale;	
	if(xx > 0xFFFF)
        {
          Raw += 3*(xx >> 16);
	  xx &= 0xFFFF;
	}
      }
    }
    else
    {
      Raster1DAbstract *pRas1D = pRas->GetRowRaster((y*YScale)>>16);
      if(pRas1D==NULL) continue;
      for(int x=0; x<width; ++x, ++p)
      {
        unsigned val = pRas1D->GetValue1D((x*XScale)>>16);
        p.Red() = val & 0xFF;
        p.Green() = (val >> 8) & 0xFF;
        p.Blue() = (val >> 16) & 0xFF;
      }
    }

    p = rowStart;
    p.OffsetY(data, 1);
  }
  //T1 = GetInterval() - T1;
  return pBitmap;
}


/**  Paints the graphs. 
 *  @param[in] event Unused event */
void CameraPreview::OnPaint(wxPaintEvent &WXUNUSED(event))
{   
int width;
int height;

  GetClientSize(&width, &height);

  Tx.X.CanvasStart = width - SHIFT_PLATE_POS;
  Tx.X.CanvasEnd = 0 + SHIFT_PLATE_POS;  
  Tx.Y.CanvasStart = 0 + SHIFT_PLATE_POS;
  Tx.Y.CanvasEnd = height - SHIFT_PLATE_POS;

  if(bit!=NULL && !ImageChanged)
  {
    if(bit->GetWidth()!=width || bit->GetHeight()!=height) ImageChanged=true;
  }

  if(ImageChanged)
  {
    Raster2DAbstract *pRas = (filter==0) ? ImageBuf : FilterBuf;
    bit = Raster2Bitmap(pRas, bit, width, height);
    ImageChanged = false;
  }

  if(ViewMode>GRAPH && bit!=NULL)
    {
    wxPaintDC dc(this);		//do not use buffers, because we have a bitmap only
    dc.DrawBitmap(*bit, 0, 0);
    videoCircle->Paint(dc);     
    }
  else
    {        
    wxBufferedPaintDC dc(this);

    dc.SetBackground(wxColour("BLACK"));
    /*if(!MemoryIsInitialised)*/ dc.Clear();    

//    DrawGraphLines(dc);
    GrX.Paint(dc);  
    GrY.Paint(dc);  
//    targetCircle->Paint(dc);
    videoCircle->Paint(dc);        
    }
}




/**
 *  Handles the mouse events and draws the desired target position.
 *  @param[in] event Used event */
void CameraPreview::OnMouseMove(wxMouseEvent& event)
{
/*
    if (event.LeftDown())
    {   // We tentatively start dragging, but wait for
        // mouse movement before dragging properly.
        m_dragMode = TEST_DRAG_START;
        dragStartPos = event.GetPosition();
    }
    else if (event.LeftUp() && m_dragMode != TEST_DRAG_NONE)
    {   // Finish dragging
        m_dragMode = TEST_DRAG_NONE;

        wxClientDC dcClient(this);
        wxSize size = dcClient.GetSize(); 
        wxBufferedDC dc;  
        dc.Init(&dcClient, size);
        dragStartPos = event.GetPosition();
                
        double val;
        val=Tx.X.Canvas2Log(dragStartPos.x);
	if(val<-1) val=-1;
        if(val>1) val=1;
	MRT.ReqX.SetDouble(val);
	val=Tx.Y.Canvas2Log(dragStartPos.y);
	if(val<-1) val=-1;
        if(val>1) val=1;
        MRT.ReqY.SetDouble(val);
        
        DrawGraphLines(dc);
        GrX.Paint(dc);
        GrY.Paint(dc);        
        targetCircle->Paint(dc);
        videoCircle->Paint(dc);
    }

    else if (event.Dragging() && m_dragMode != TEST_DRAG_NONE)
    {
        if (m_dragMode == TEST_DRAG_START)
        {
            // We will start dragging if we've moved beyond a couple of pixels
            wxPoint pos = event.GetPosition();
            int tolerance = 2;
            int dx = abs(pos.x - dragStartPos.x);
            int dy = abs(pos.y - dragStartPos.y);
            if (dx <= tolerance && dy <= tolerance)
                return;
            
            // Start the drag.
//          m_dragMode = TEST_DRAG_DRAGGING;
                        
            wxClientDC dcClient(this);
            wxSize size = dcClient.GetSize(); 
            wxBufferedDC dc;  
            dc.Init(&dcClient, size);
                       
            wxCoord w, h;
            dc.GetSize(&w, &h);
            wxPoint p = wxPoint(w, h);
            
	    double val;
            val=Tx.X.Canvas2Log(pos.x);
	    if(val<-1) val=-1;
            if(val>1) val=1;
	    MRT.ReqX.SetDouble(val);
	    val=Tx.Y.Canvas2Log(pos.y);
	    if(val<-1) val=-1;
            if(val>1) val=1;
            MRT.ReqY.SetDouble(val);
                        
            DrawGraphLines(dc);            
            GrX.Paint(dc);  
            GrY.Paint(dc);  

	    targetCircle->Paint(dc);
	    videoCircle->Paint(dc);
        }
       // else if (m_dragMode == TEST_DRAG_DRAGGING)
       // {
            // We're currently dragging. See if we're over another shape.
       // }
    }    
*/
}


/**
 *  Sets background appearance.
 *
 *  @see wxColour is the reference class for any further question about the parameter of the method.
 *  @param[in] color wxColour value
 */
void CameraPreview::SetBackground(const wxColour &color)
{   
    if(color.Ok())
        backgroundColor = color;    
    else     
        backgroundColor = wxColor(DEFAULT_BGROUND); 
}


/**
 *  Sets the background of the panel. Unless it is assigned with
 *  different value, it uses the default value, i.e., "RED"
 *
 *  @param[in] devContext A device context is onto which graphics and text can be drawn */
void CameraPreview::SetBackGColorProp(wxDC &devContent)
{  
    // the color of frame line of outter rectangle
  devContent.SetPen( wxPen( wxT("BLUE"), 3, wxSOLID) );
  devContent.SetBrush( wxBrush(backgroundColor) );
}


/**
 *  Sets the target circle appearance.
 *
 *  @see wxColour is the reference class for any further question about the parameter of the method.
 *  @param[in] color wxColour value  */
void CameraPreview::SetTCircleColor(const wxColour &color)
{   
  if(color.Ok())
     targetCircle->color = color;    
  else     
     targetCircle->color = wxColor(DEFAULT_TCIRCLE_COLOR); 
}


/**
 *  Sets the video circle appearance.
 *
 *  @see wxColour is the reference class for any further question about the parameter of the method.
 *  @param[in] color wxColour value
 */
void CameraPreview::SetVCircleColor(const wxColour &color)
{   
  if(color.Ok())
    videoCircle->color = color;    
  else     
    videoCircle->color = wxColor(DEFAULT_VCIRCLE_COLOR); 
}


/**
 *  Sets grid line appearence.
 *
 *  @see wxPen is the reference class  for any further question about the parameter of the method. 
 *  @param[in] pen wxPen */
void CameraPreview::SetGridLine(const wxPen &pen)
{
  if(pen.Ok())
    GrX.LinePen=GrY.LinePen = pen;
  else
    GrX.LinePen=GrY.LinePen = wxPen(DEFAULT_GRID_LINE); 
}


/** Exchange sz bytes between block pointed by ptr1 and ptr2. */
/* void flipdata(BYTE *ptr1, BYTE *ptr2, unsigned sz)
{
DWORD buff;
  while(sz>=4)
  {
    buff = *(DWORD*)ptr1;
    *(DWORD*)ptr1 = *(DWORD*)ptr2;
    *(DWORD*)ptr2 = buff;
    ptr1+=4;
    ptr2+=4;
    sz-=4;
  }
  while(sz>=1)
  {
    buff = *ptr1;
    *ptr1 = *ptr2;
    *ptr2 = buff;
    ptr1+=1;
    ptr2+=1;
    sz-=1;
  }
} */


/** Captures image according to a ViewMode set. */
bool CameraPreview::CaptureImage(void)
{
  if(CT->Camera==NULL) 
  {        //nothing to capture	 
    return false;  /// \todo: here could be assert, but I prefer to silently return
  }  

  switch(ViewMode)
  {
    case RAW_PIC:         CT->ReadImageAs(ImageBuf, 1|(rectangle?0x80:0), vflip, hflip); break;
    case THRESHOLDED_PIC: CT->ReadImageAs(ImageBuf, 2|(rectangle?0x80:0), vflip, hflip); break;
    case RAW_COLOR:	  CT->ReadImageAs(ImageBuf, 0|(rectangle?0x80:0), vflip, hflip); break;
    default: return false;  //no capture needed
  }
  
  const int width = ImageBuf->Size1D;
  const int height = ImageBuf->Size2D;
  if(width==0 || height==0) return true;  

  if(filter==0)
  {
    FilterBuf->Erase();
  }
  else
  {
    bool empty = FilterBuf->Data2D==NULL;
    FilterBuf->Allocate2D(width,height);

    unsigned char *ImgData    = (unsigned char *)ImageBuf->GetRow(0);
    unsigned char *FilterData = (unsigned char *)FilterBuf->GetRow(0);

    if(empty)
      memcpy(FilterData,ImgData,width*height*3);
    else
    {      
      for(long i=width*height*3; i>0; i--)
      {
        *FilterData = ((int)*ImgData + (int)*FilterData)/2;
        FilterData++;
	ImgData++;
      }
    }
  }
  ImageChanged = true;

  return true;
}


void CameraPreview::HookFrame(wxFrame *pFrame2Hook)
{
  if(CT->Camera==NULL) return;
  CT->Camera->HookFrame(pFrame2Hook);
}


int CameraPreview::getFrameCounter(void)
{
  if(CT==NULL || CT->Camera==NULL) return 0; 
  const int i=CT->Camera->FrameCounter; 
  CT->Camera->FrameCounter=0; 
return i;
}